class BSTNode {
    BSTNode parent = null;
    BSTNode left = null;
    BSTNode right = null;
    int key;
    int height;

    public void updateHeight()
    {
        if (left == null)
        {
            height = 0;
        }
        else if (left.height > right.height)
        {
            height = left.height + 1;
        }
        else
        {
            height = right.height + 1;
        }
    }
    public void printAllNodes()
    {
        printNodesRecurse(this);
    }
    public void printNodesRecurse(BSTNode current)
    {
        if (current.left != null)
        {
            printNodesRecurse(current.left);
            if (current.parent != null) {
                System.out.println("Node with key " + current.key + ", parent has key "
                        + current.parent.key + ", current height is " + current.height);
            } else
            {
                System.out.println("Node with key " + current.key +
                        ", this is the root, current height is " + current.height);
            }

            printNodesRecurse(current.right);
        }

    }
}

public class Main {

    public static void main(String[] args)
    {
        /*Since we're manually building a small snippet of a BST, these
         * are all the nodes we'll need, named for the role they
         * play in each case rather than their specific locations.
         * Note that the "externals" are more like stand-ins for subtrees
         * in this case, as their actual identities should be irrelevant to
         * the rotation. They're essentially bystanders.*/
        BSTNode unbalancedRoot = new BSTNode();
        BSTNode rootHigherChild = new BSTNode();
        BSTNode nextHigherChild = new BSTNode();
        BSTNode rootLowerChild = new BSTNode();
        BSTNode nextLowerChild = new BSTNode();
        BSTNode externalRootLowerChildLeft = new BSTNode();
        BSTNode externalRootLowerChildRight = new BSTNode();
        BSTNode externalNextHigherChildLeft = new BSTNode();
        BSTNode externalNextHigherChildRight = new BSTNode();
        BSTNode externalNextLowerChildLeft = new BSTNode();
        BSTNode externalNextLowerChildRight = new BSTNode();

        //Our root node, which in this case has become unbalanced.
        unbalancedRoot.parent = null;
        unbalancedRoot.left = rootLowerChild;
        unbalancedRoot.right = rootHigherChild;
        unbalancedRoot.key = 5;
        unbalancedRoot.height = 4;
        //The root node's right child, with a higher key, as well as a height that's two higher than its sibling, which
        //is the cause of their parent's imbalance
        rootHigherChild.parent = unbalancedRoot;
        rootHigherChild.left = nextLowerChild;
        rootHigherChild.right = nextHigherChild;
        rootHigherChild.key = 10;
        rootHigherChild.height = 3;
        //The right child of the root node's right child, with the highest key of the internal nodes we're looking at,
        //and a higher height than its sibling - it may have been an insertion into its subtree that unbalanced the root
        nextHigherChild.parent = rootHigherChild;
        nextHigherChild.left = externalNextHigherChildLeft;
        nextHigherChild.right = externalNextHigherChildRight;
        nextHigherChild.key = 15;
        nextHigherChild.height = 2;
        //The left child of the root, with a smaller key than root and a height two lower than their sibling.
        rootLowerChild.parent = unbalancedRoot;
        rootLowerChild.left = externalRootLowerChildLeft;
        rootLowerChild.right = externalRootLowerChildRight;
        rootLowerChild.key = 1;
        rootLowerChild.height = 1;
        //The left child of the root node's right child, with a lower key than their parent and a height one lower than
        //their sibling's.
        nextLowerChild.parent = rootHigherChild;
        nextLowerChild.left = externalNextLowerChildLeft;
        nextLowerChild.right = externalNextLowerChildRight;
        nextLowerChild.key = 13;
        nextLowerChild.height = 1;

        externalRootLowerChildLeft.height = 0;
        externalRootLowerChildRight.height = 0;
        externalNextHigherChildLeft.height = 0;
        externalNextHigherChildRight.height = 0;
        externalNextLowerChildLeft.height = 0;
        //This external node's height has been artificially increased to 1, to represent a larger subtree
        //that's indirectly causing the imbalance at the root, likely as the result of an insertion down this way.
        externalNextLowerChildRight.height = 1;

        System.out.println("Pre restructure: ");
        unbalancedRoot.printAllNodes();


        //This is where you need to call your function from the TrinodeRestructurer class and restructure the tree
        TrinodeRestructurer.restructure(unbalancedRoot);

        System.out.println("\nPost restructure: ");
        unbalancedRoot.printAllNodes();
    }

}
